package se.sics.streamingclassifier

@SerialVersionUID(1729)
class PredictionResults(private var _numClasses: Int, var confusionMatrix: Array[Array[Long]]) extends Serializable {
  var testWindowId: Long = _
  var totalTrainingCount: Long = _
  var comments: String = ""

  def this(numClasses: Int) = {
    this(numClasses, (0 until numClasses).map(_ => Array.fill(numClasses)(0L)).toArray)
  }

  def this() = {
    this(0)
  }

  // We override the setter for numClasses so that the matrix gets updated when numClasses changes.
  // We'll get indexing issues otherwise.
  def numClasses= _numClasses
  def numClasses_= (newNumClasses: Int) {
    //When the numClasses change we need to update the count matrix. For now, we just initialize it to zero
    _numClasses = newNumClasses
    confusionMatrix = (0 until newNumClasses).map(_ => Array.fill(newNumClasses)(0L)).toArray
  }

  /**
   * Create a new ConfusionMatrix which is the result of observing a new labeled prediction
   * @param labeledPrediction: The new observed LabeledPrediction
   * @return A new ConfusionMatrix with the observation added to the counts.
   */
  def addPrediction(labeledPrediction: LabeledPrediction): PredictionResults = labeledPrediction match {
    case LabeledPrediction(trueLabel, predictedLabel) =>
      val updatedCount = confusionMatrix.map(_.clone())
      val row = updatedCount(trueLabel)
      row(predictedLabel) += 1
      new PredictionResults(numClasses, updatedCount)
  }

  def calculateTotalCount(): Long = {
    confusionMatrix.map(_.sum).sum
  }

  def calculateMissCount(): Long = {
    confusionMatrix.zipWithIndex.map {
      case (classPredictions, trueClass) =>
        classPredictions.zipWithIndex.
          filter { case (count, predictedClass) => predictedClass != trueClass}.
          map(_._1).
          sum
    }.sum
  }

  override def toString = {
    var output = new StringBuilder("")

    val totalCount: Long = calculateTotalCount()
    val missClassCount: Long = calculateMissCount()
    val missClassRatio: Double = missClassCount.toDouble / totalCount

    output ++= s"Test window number: $testWindowId\n"
    output ++= s"Seen training examples: $totalTrainingCount\n"
    output ++= f"Missclass ratio (incorrect predictions / total observations): $missClassRatio%.05f"  + "\n"
    output ++= "Confusion matrix"  + "\n"
    output ++= "              Predicted label" + "\n"
    output ++= "              _" + confusionMatrix.indices.map(x => f"$x%5d").mkString(" ").replace(' ', '_') + "\n"

    // We treat the first row differently
    output ++= f"True label ${0}%2d| "
    output ++= confusionMatrix(0).map(x => f"$x%5d").mkString(" ") + "\n"

    for (trueLabel <- 1 until confusionMatrix.length) {
      output ++= f"           $trueLabel%2d| "
      val classPredictions: Array[Long] = confusionMatrix(trueLabel)
      output ++= classPredictions.map(x => f"$x%5d").mkString(" ")  + "\n"
    }

    if (comments.nonEmpty) output ++= comments + "\n"

    val maxRow = maxRowLength(output.toString())
    val border = "=" * maxRow

    border + "\n" + output.toString() + border + "\n"
  }

  /**
   * Returns the maximum row length of the given newline-separated lines
   * @param lines: A string with newline-separated lines.
   * @return The length of the longest line.
   */
  def maxRowLength(lines: String): Int = {
    lines.lines.map(line => line.length).max
  }
}
