package se.sics.streamingclassifier

import java.io.{BufferedWriter, FileWriter, File}

import org.apache.flink.api.common.functions.FoldFunction

import org.apache.flink.api.java.utils.ParameterTool

import org.apache.flink.streaming.api.scala._
import org.apache.flink.streaming.api.scala.createTypeInformation
import org.apache.flink.streaming.api.windowing.helper.Count

import scala.io.Source

object StreamingClassifier {
  def main(args: Array[String]) {
    val parameters = ParameterTool.fromArgs(args)

    val trainingFile = parameters.getRequired("training-data")
    val testFile = parameters.getRequired("test-data")
    val outputFile = parameters.get("output", "prediction_output.txt")
    val delimiter = parameters.get("delimiter", ",")

    val numFeatures = getNumColumns(trainingFile, delimiter) - 1  // -1 since one column is the class label
    val classColumn = parameters.getLong("class-column", 0).toInt

    // Having more classes then there actually are is no problem. They will have zero probability of being chosen.
    // We choose a default value large enought for this excercise.
    val numClasses = parameters.getLong("num-classes", 10).toInt
    val windowSize = parameters.getLong("window-size", 1000).toInt
    val strideSize = parameters.getLong("stride-size", 100).toInt

    val classifier = new NaiveBayesModel(numClasses, numFeatures)
    println("Starting a new training with the following parameters: ")
    println(s"--training-data: $trainingFile")
    println(s"--test-data: $testFile")
    println(s"--num-classes: $numClasses")
    println(s"--class-column: $classColumn")
    println(s"--window-size: $windowSize")
    println(s"--stride-size: $strideSize")
    println(s"--delimiter: $delimiter")
    println(s"Number of features: $numFeatures")

    testModelStreaming(classifier, trainingFile, testFile, outputFile, delimiter = delimiter, classColumn = classColumn, windowSize=windowSize, strideSize = strideSize)
    //testModelLocal(classifier, trainingFile, testFile, outputFile, delimiter = delimiter, classColumn = classColumn, windowSize=windowSize, strideSize = strideSize)
  }

  /**
   * Runs the model on a combined stream of training and test data. Makes prediction on sliding windows of data
   * from the test stream.
   * @param model: A ClassificationModel instance to train.
   * @param trainingFile: A path to a file with labeled samples to use for training.
   * @param testFile: A path to a file with labeled samples to use for evaluation.
   * @param outputPath: A path to a file which output will be written to.
   * @param delimiter: Delimiter to use when splitting the file.
   * @param classColumn: Which column index should be used for the class label.
   * @param windowSize: The size of windows to calculate performance on.
   * @param strideSize: The stride size, how many samples to forward until the next window.
   */
  def testModelStreaming(model: ClassificationModel,
                         trainingFile: String,
                         testFile: String,
                         outputPath: String,
                         delimiter: String = ",",
                         classColumn: Int = 0,
                         windowSize: Int = 10000,
                         strideSize: Int = 1000) = {
    val env: StreamExecutionEnvironment = StreamExecutionEnvironment.getExecutionEnvironment

    val trainingStream: DataStream[LabeledSample] = env.readTextFile(trainingFile).flatMap(new Labeler(delimiter, classColumn))
    val testStream: DataStream[LabeledSample] = env.readTextFile(testFile).flatMap(new Labeler(delimiter, classColumn))

    // We join the training and test streams. The Training stream is used to update the model, the test stream is used
    // To produce predictions. The output stream only contains predictions from the test stream
    val windowedTestData = testStream.window(Count.of(windowSize)).every(Count.of(strideSize)).getDiscretizedStream()
    val confusionStream: DataStream[PredictionResults] = trainingStream.connect(windowedTestData).flatMap(new ClassifierOperator(model))

    // Now we write out the confusion matrix stream
    confusionStream.writeAsText(outputPath)
    env.execute("Naive Bayes stream classification")
  }

  /**
   * Runs the model one sample at a time serialized on a collected array of the training and test dataset.
   * @param model: A ClassificationModel instance to train.
   * @param trainingFile: A path to a file with labeled samples to use for training.
   * @param testFile: A path to a file with labeled samples to use for evaluation.
   * @param outputPath: A path to a file which output will be written to.
   * @param delimiter: Delimiter to use when splitting the file.
   * @param classColumn: Which column index should be used for the class label.
   * @param windowSize: The size of windows to calculate performance on.
   * @param strideSize: The stride size, how many samples to forward until the next window.
   */
  def testModelLocal(model: ClassificationModel,
                         trainingFile: String,
                         testFile: String,
                         outputPath: String,
                         delimiter: String = ",",
                         classColumn: Int = 0,
                         windowSize: Int = 1000,
                         strideSize: Int = 10) = {
    val trainingData = Source.fromFile(trainingFile).getLines().drop(1).map(row => makeLabeled(row, delimiter, classColumn))
    val testData = Source.fromFile(testFile).getLines().drop(1).map(row => makeLabeled(row, delimiter, classColumn))
    val windowedTrainingData = trainingData.sliding(10, 10)
    val windowedTestData = testData.sliding(windowSize, strideSize)

    var totalTrainingCount = 0
    val performances = windowedTrainingData.zip(windowedTestData).zipWithIndex.map {
      case ((trainingWindow, testWindow), windowCount) =>
        trainingWindow.foreach(model.update)
        totalTrainingCount += trainingWindow.length

        val predictions = testWindow.map {
          case LabeledSample(classLabel, features) =>
            LabeledPrediction(classLabel, model.predict(UnlabeledSample(features)))
        }

        val initialMatrix = new PredictionResults(model.numClasses)
        def folder(previousMatrix: PredictionResults, prediction: LabeledPrediction): PredictionResults =
          previousMatrix.addPrediction(prediction)

        val confusionMatrix: PredictionResults = predictions.foldLeft(initialMatrix)(folder)
        confusionMatrix.totalTrainingCount = totalTrainingCount
        confusionMatrix.testWindowId = windowCount
        confusionMatrix.comments = model.toString
        confusionMatrix
    }

    val file = new File(outputPath)
    val bw = new BufferedWriter(new FileWriter(file))
    performances.foreach(performance => bw.write(performance.toString))
    bw.close()
  }
}
