package se.sics

import org.apache.flink.api.common.functions.FlatMapFunction
import org.apache.flink.util.Collector

import scala.io.Source

import org.apache.flink.api.scala.{DataSet, createTypeInformation}

package object streamingclassifier {
  // Just some clarifying type definitions
  type Feature = Double
  type ClassLabel = Int
  type Mean = Double
  type Variance = Double

  case class LabeledPrediction(label: ClassLabel, prediction: ClassLabel)

  case class LabeledSample(label: ClassLabel, features: Array[Feature]) {
    override def toString = {
      s"LabeledSample: class($label), features: ${features.mkString(",")}"
    }
  }
  case class UnlabeledSample(features: Array[Feature])

  def pointwiseMul(x: Array[Double], y: Array[Double]): Array[Double] = {
    x.zip(y).map({ case (xi, yi) => xi * yi })
  }

  def pointwiseSum(x: Array[Double], y: Array[Double]): Array[Double] = {
    x.zip(y).map({ case (xi, yi) => xi + yi })
  }

  def pointwiseSub(x: Array[Double], y: Array[Double]): Array[Double] = {
    x.zip(y).map({ case (xi, yi) => xi - yi })
  }

  def scalarMul(k: Double, x: Array[Double]): Array[Double] = {
    x.map(k * _)
  }

  def scalarDiv(k: Double, x: Array[Double]): Array[Double] = {
    val reciprocal = 1/k
    scalarMul(reciprocal, x)
  }

  def getNumColumns(filePath: String, delimiter: String = ","): Int = {
    val firstLines = Source.fromFile(filePath).getLines().take(10)
    // We gather count the columns for a couple of lines
    val columns = firstLines.map(_.split(delimiter).length).toArray
    // We tally the column lengths. We expect all of them to have the same length, or at least the majority.
    //
    val tallied: Array[(Int, Int)] = columns.groupBy(identity).map(t => (t._2.length, t._1)).toArray
    val (maxCount, numCols) = tallied.max
    numCols
  }

  /**
   * Calculate the missclassification ratio.
   * @param predictions
   */
  def getMissclassRatio(predictions: DataSet[(ClassLabel, ClassLabel)]): Double = {

    val counts: (Int, Int) = predictions.
      // First we create count touples, the first element keeps track of the total number of examples, the second the
      // number of correct predictions.
      map(prediction => (1, if (prediction._1 == prediction._2) 1 else 0)).
      // Now we just sum up the counts
      reduce((a, b) => (a._1 + b._1, a._2 + b._2)).
      collect().head
    // The missclass ratio is 1 - correctCount/totalCount
    1 - counts._2.toDouble / counts._1
  }

  /**
   * Calculates a confusion matrix for the given predictions.
   * @param predictions
   * @return A nested array tagged with class labels. The outer array is tuples with the true class and the second is
   *         a distribution over prediction for that class.
   */
  def getConfusionMatrix(predictions: DataSet[(ClassLabel, ClassLabel)]): Array[Array[Int]] = {
    val perClassCount = predictions.
      // First we make a triple. The first element is the true label, the second the predicted and the third used to count the occurances
      map(prediction => (prediction._1, prediction._2, 1)).
      // First we group by the true class label
      groupBy(0, 1).
      reduce((a, b) => (a._1, a._2, a._3 + b._3)).collect().toArray

    // We get the class number by looking at the counts
    val numClasses = {
      var maxclass = 0
      for (triple <- perClassCount) {
        triple match { case (trueLabel, predictedLabel, count) =>
          if (trueLabel > maxclass) maxclass = trueLabel
          else if (predictedLabel > maxclass) maxclass = predictedLabel
        }
      }
      maxclass + 1 // We need to add 1, since maxClass is 0-indexed
    }

    val confusionMatrix: Array[Array[Int]] = Array.ofDim[Int](numClasses, numClasses)
    // Now we put the counts in a matrix

    for (triple <- perClassCount) {
      triple match { case (trueLabel, predictedLabel, count) =>
        confusionMatrix(trueLabel)(predictedLabel) = count
      }
    }
    confusionMatrix
  }

  def printMetrics(predictions: DataSet[(ClassLabel, ClassLabel)]) = {
    val missClassRatio = getMissclassRatio(predictions)
    val confusionMatrix = getConfusionMatrix(predictions)

    println(f"Missclass ratio (incorrect predictions / total observations): $missClassRatio%.05f")
    println("Confusion matrix")
    println("               Predicted label")
    println("               " + confusionMatrix.indices.map(x => f"$x%5d").mkString(" "))

    // We treat the first row differently
    print(f"True label ${0}%2d| ")
    println(confusionMatrix(0).map(x => f"$x%5d").mkString(" "))

    for (trueLabel <- 1 until confusionMatrix.length) {
      print(f"           $trueLabel%2d| ")
      val classPredictions: Array[Int] = confusionMatrix(trueLabel)
      println(classPredictions.map(x => f"$x%5d").mkString(" "))
    }
  }


  /**
   * Class used to convert a stream of rows from a file into LabeledSamples
   * @param delimiter: String delimiter to use for splitting the fields of the row.
   * @param classColumn: Which column (0 indexed) is the class column.
   */
  class Labeler(val delimiter:String = ",", val classColumn: Int = 0) extends FlatMapFunction[String, LabeledSample] {
    /**
     * Convert a row to a LabeledSample.
     * @param row: The row to convert. If the row can't be converted it is silently ignored.
     * @param collector: The Collector object to write samples to.
     */
    override def flatMap(row: String, collector: Collector[LabeledSample]): Unit = {
      scala.util.control.Exception.ignoring(classOf[NumberFormatException]) {

        collector.collect(makeLabeled(row, delimiter, classColumn))
      }
    }
  }

  def makeLabeled(row: String, delimiter: String = ",", classColumn: Int = 0): LabeledSample = {
    val tokens = row.split(delimiter).toBuffer  // Buffers have the method remove which we'll use to get all elements but the class label
    val classLabel = tokens.remove(classColumn).toDouble.toInt
    val features: Array[Double] = tokens.map(_.toDouble).toArray
    LabeledSample(classLabel, features)
  }
}
